@@ -1,4 +1,4 @@
-import os
+import os,sys
 import warnings
 import itertools
 from nextnanopy.utils.formatting import text_to_lines, lines_to_text
@@ -95,6 +95,7 @@ class InputFileTemplate(object):
     def __init__(self, fullpath=None, configpath=None):
         self.raw_lines = []
         self.variables = DictList()
+        self.content = None
         self.fullpath = fullpath
         self.product = 'not valid'
         if fullpath is not None:
@@ -116,6 +117,10 @@ class InputFileTemplate(object):
         self.validate()
         self.load_variables()
 
+    @property
+    def raw_text(self):
+        return str(lines_to_text(*self.raw_lines))
+
     @property
     def lines(self):
         new_lines = list(self.raw_lines)
@@ -202,14 +207,16 @@ class InputFileTemplate(object):
         self.find_product()
         self.validate()
         self.load_variables()
+        self.load_content()
 
     def find_product(self):
-        self.product = defaults.input_text_type(self.text)
+        self.product = defaults.input_text_type(self.raw_text)
         return self.product
 
     def validate(self):
         if self.product not in defaults.products:
-            raise ValueError(f'Not valid input file')
+            self.product = 'Not valid'
+            #raise ValueError(f'Not valid input file')
 
     @save_message
     def save(self, fullpath=None, overwrite=False, automkdir=True):
@@ -236,7 +243,7 @@ class InputFileTemplate(object):
         return self.fullpath
 
     @execute_message
-    def execute(self, **kwargs):
+    def execute(self, show_log = True, convergenceCheck = False, **kwargs):
         """
         Execute the input file located at .fullpath
         Individual kwargs can be passed like 'license' or 'database'
@@ -244,28 +251,88 @@ class InputFileTemplate(object):
 
         Parameters
         ----------
-        exe : str, optional
-            path to executable
-        license : str, optional
-            path to license file
-        database : str, optional
-            path to database file
-        outputdirectory : str, optional
-            path where to save the simulated data
+        show_log : bool, optional
+            if False, suppress the simulation log
+            (default is True)
+        convergenceCheck : bool, optional
+            if True, check convergence of the simulation
+            (default is False)
 
-        Other parameters can be used depending on the nextnano product.
-        For example, 'threads' is accepted for nextnano++.
-        Please, see the documentation of the command line arguments for each nextnano product
-        in the website (https://www.nextnano.com/)
+        kwargs may contain:
+            exe : str, optional
+                path to executable
+            license : str, optional
+                path to license file
+            database : str, optional
+                path to database file
+            outputdirectory : str, optional
+                path where to save the simulated data
+        and other parameters depending on the nextnano product.
+        For example, 'threads' is accepted by nextnano++.
+        See the documentation of the command line arguments of each nextnano product
+        on the online Manual (https://www.nextnano.de/manual/).
         """
 
         cmd_kwargs = dict(self.default_command_args)
         cmd_kwargs.update(kwargs)
         cmd_kwargs['inputfile'] = self.fullpath
-        info = cmd_execute(**cmd_kwargs)
+        info = cmd_execute(show_log=show_log, **cmd_kwargs)
         self.execute_info = info
+        if convergenceCheck:
+            self.check_convergence()
         return info
 
+    def check_convergence(self): 
+        if self.product == 'nextnano.MSB':
+            raise NotImplementedError('Convergence check has not yet implemented for nextnano.MSB!')
+            
+        log = self.execute_info['logfile']
+        try:
+            if self.product == 'nextnano3':
+                with open(log, 'r') as file:
+                    for line in file:
+                        if 'Exiting iteration and terminating simulation' in line or 'Program was terminated using a soft kill' in line or 'Terminating immediately' in line:
+                            raise RuntimeError(f'\nSimulation got terminated! Check the log:\n{log}')
+                        elif 'Maximum number of iterations exceeded' in line:
+                            raise RuntimeError(f'\nSimulation did not converge! Check the log:\n{log}')
+            elif self.product == 'nextnano++':
+                with open(log, 'r') as file:
+                    for line in file:
+                        if 'Terminating program' in line:
+                            raise RuntimeError(f'\nSimulation got terminated! Check the log:\n{log}')
+                        elif 'Maximum number of iterations exceeded' in line:
+                            raise RuntimeError(f'\nSimulation did not converge! Check the log:\n{log}')
+                        elif 'Outdated numerics library (f95library) used' in line:
+                            raise RuntimeError(f'\nOutdated numerics library (f95library) used.')
+            elif self.product == 'nextnano.NEGF':   # nextnano.NEGF reports convergence at every voltage and temperature sweep.
+                with open(log, 'r') as file:
+                    for line in file:
+                        if 'Simulation has NOT CONVERGED' in line:
+                            raise RuntimeError(f'\nSimulation has diverged! Check the log:\n{log}')
+                        elif 'Simulation has partially converged' in line:
+                            print(f'\nWARNING: check_convergence(): Simulation did not fully converge.')
+                        elif 'Terminating program!' in line:
+                            raise RuntimeError(f'\nSimulation got terminated! Check the log:\n{log}')
+        except FileNotFoundError:
+            print(f'Log file {log} not found!')
+            raise
+        except RuntimeError as e:
+            print(e)
+            pause = True
+            while pause:
+                answer = input('Do you nevertheless want to continue? [y/n]: ')
+                if answer == 'y' or answer == 'yes': 
+                    pause = False
+                    return
+                elif answer == 'n' or answer == 'no': 
+                    pause = False
+                    raise RuntimeError('Nextnanopy terminated.')
+                else: 
+                    print('Invalid input.')
+                    continue
+        else:
+            return
+
     def clear(self):
         self.raw_lines = []
         self.variables = DictList()
@@ -283,6 +350,9 @@ class InputFileTemplate(object):
     def load_variables(self):
         pass
 
+    def load_content(self):
+        pass
+
     def get_variable(self, name):
         """
         Equivalent to self.variables[name]
@@ -301,7 +371,7 @@ class InputFileTemplate(object):
             raise KeyError(f'{name} is not a valid variable.')
         return self.variables[name]
 
-    def set_variable(self, name, value=None, comment=None):
+    def set_variable(self, name, value=None, comment=None, unit = None):
         """
         Change the value and/or the comment of self.variable[name]
 
@@ -315,6 +385,9 @@ class InputFileTemplate(object):
         comment : not defined, optional
             Equivalent to self.variables[name].comment = comment (default is None)
             If it is None, it won't change it
+        unit : not defined, optional
+            Equivalent to self.variables[name].unit = unit (default is None)
+            If it is None, it won't change it
         """
 
         var = self.get_variable(name)
@@ -322,6 +395,8 @@ class InputFileTemplate(object):
             var.value = value
         if comment is not None:
             var.comment = comment
+        if unit is not None:
+            var.unit  = unit
         return var
 
     def __getitem__(self, item):
@@ -357,6 +432,7 @@ class InputFileTemplate(object):
 
 
 class InputFile(InputFileTemplate):
+    # def __new__(cls, fullpath=None, configpath=None):
 
     def load_variables(self):
         """
@@ -368,6 +444,34 @@ class InputFile(InputFileTemplate):
         file.text = self.text
         self.variables = file.variables
 
+    def load_content(self):
+        _InputFile = defaults.get_InputFile(self.product)
+        file = _InputFile()
+        file.text = self.text
+        file.load_content()
+        self.content = file.content
+
+    @property
+    def lines(self):
+        """
+        update method to support NEGF input
+        """
+        try:
+            _InputFile = defaults.get_InputFile(self.product)
+        except ValueError:
+            _InputFile = InputFileTemplate
+
+
+        file = _InputFile()
+        #file.product = self.product
+        #file.text = self.raw_text
+        #file.variables = self.variables
+        #print(file.product)
+        #print(type(file))
+        file.variables = self.variables
+        file.raw_lines = self.raw_lines
+        return file.lines
+
 class Sweep(InputFile):
     """
         This class give a user possibility to run multiple simulations (sweep) over defined variables in the input file.
@@ -429,13 +533,32 @@ class Sweep(InputFile):
             self.input_files.append(inputfile)
 
 
-    def execute_sweep(self, delete_input_files = False, overwrite = False):
+    def execute_sweep(self, delete_input_files = False, overwrite = False, show_log = True, convergenceCheck = False):
+        """
+        Execute created input files and saves information to output folder.
+
+        Parameters
+        ----------
+        delete_input_files : bool, optional
+            if True, input_files are deleted after execution. Default is False.
+        overwrite : bool, optional
+            if True, the output overwrites the old output data. If False, execution will create a new output folder
+            (with the unique name, created by adding an integer to the foldername). Default is False.
+        show_log : bool, optional
+            if True, the simulation log is displayed in the console, while False suppresses the log. Default is True.
+            Note that the log file is always saved in the output folders regardless of this option.
+        convergenceCheck : bool, optional
+            if True, nextnanopy scans the log file of the simulation performed and check whether the solution has converged. 
+            If it did not converge, nextnanopy warns you and ask if you want to proceed with postprocessing. 
+            Note that non-converged solutions are not reliable and further calculation and/or visualization from them do not make much sense. 
+            Default is False.
+        """
         self.prepare_output(overwrite)
         output_directory = self.sweep_output_directory
         if not self.input_files:
             warnings.warn('Nothing was executed in sweep! Input files to execute were not created.')
         for inputfile in self.input_files:
-            inputfile.execute(outputdirectory = output_directory)
+            inputfile.execute(outputdirectory = output_directory, show_log = show_log, convergenceCheck = convergenceCheck)
             if delete_input_files:
                 inputfile.remove()
 
